#include <windows.h>
#include <mmsystem.h>
#include <mmdeviceapi.h>
#include <mmddk.h>
#include <strsafe.h>

//-----------------------------------------------------------
// This function gets the waveOut ID of the audio endpoint
// device that is currently assigned to the specified device
// role. The caller can use the waveOut ID to open the
// waveOut device that corresponds to the endpoint device.
//-----------------------------------------------------------
#define EXIT_ON_ERROR(hres)  \
              if (FAILED(hres)) \
				{ char achMessage[255]; \
				sprintf(achMessage, "HRES is %u", (hres)); \
				MessageBox(HWND_DESKTOP, achMessage, "EXIT_ON_ERROR message", 0); \
				goto Exit; }
#define SAFE_RELEASE(punk)  \
              if ((punk) != NULL)  \
                { (punk)->Release(); (punk) = NULL; }

extern "C" {
unsigned int GetWaveDeviceID()
{
    HRESULT  hr;
    IMMDeviceEnumerator  *pEnumerator = NULL;
    IMMDevice  *pDevice = NULL;
    WCHAR  *pstrEndpointIdKey = NULL;
    WCHAR  *pstrEndpointId = NULL;
	static int nDeviceID = -1;

	// If we've already figured this out, return it
	if (nDeviceID != -1) return nDeviceID;

	// If this is not Windows Vista, return the handle from WAVE_MAPPER
	OSVERSIONINFOEX osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	GetVersionEx((LPOSVERSIONINFO) &osvi);
	if (osvi.dwMajorVersion < 6)
	{
		HWAVEOUT hwo;
		WAVEFORMATEX fmt;
		fmt.wFormatTag = WAVE_FORMAT_PCM;      
		fmt.nChannels = 2;
		fmt.wBitsPerSample = 16;
		fmt.nBlockAlign = fmt.nChannels * fmt.wBitsPerSample / 8;
		fmt.nSamplesPerSec = 44100;
		fmt.nAvgBytesPerSec = fmt.nSamplesPerSec * fmt.nBlockAlign;
		fmt.cbSize = 0;
		return nDeviceID = (unsigned int) waveOutOpen(&hwo, WAVE_MAPPER, &fmt, 0, 0, WAVE_FORMAT_QUERY);
	}

	// Create an audio endpoint device enumerator.
	hr = CoCreateInstance(__uuidof(MMDeviceEnumerator),
						  NULL, CLSCTX_INPROC_SERVER,
						  __uuidof(IMMDeviceEnumerator),
						  (void**)&pEnumerator);
	EXIT_ON_ERROR(hr)

	// Get the audio endpoint device that the user has
	// assigned to the specified device role.
	hr = pEnumerator->GetDefaultAudioEndpoint(eRender, eMultimedia,
											  &pDevice);
	EXIT_ON_ERROR(hr)

	// Get the endpoint ID string of the audio endpoint device.
	hr = pDevice->GetId(&pstrEndpointIdKey);
	EXIT_ON_ERROR(hr)

	// Get the size of the endpoint ID string.
	size_t  cbEndpointIdKey;
	hr = StringCbLengthW(pstrEndpointIdKey,
						STRSAFE_MAX_CCH * sizeof(WCHAR),
						&cbEndpointIdKey);
	EXIT_ON_ERROR(hr)

	// Include terminating null in string size.
	cbEndpointIdKey += sizeof(WCHAR);

	// Allocate a buffer for a second string of the same size.
	pstrEndpointId = (WCHAR*)CoTaskMemAlloc(cbEndpointIdKey);
	if (pstrEndpointId == NULL)
	{
		EXIT_ON_ERROR(hr = E_OUTOFMEMORY)
	}

	// Each for-loop iteration below compares the endpoint ID
	// string of the audio endpoint device to the endpoint ID
	// string of an enumerated waveOut device. If the strings
	// match, then we've found the waveOut device that is
	// assigned to the specified device role.
	int waveOutId;
	int cWaveOutDevices = waveOutGetNumDevs();

	for (waveOutId = 0; waveOutId < cWaveOutDevices; waveOutId++)
	{
		MMRESULT  mmr;
		size_t  cbEndpointId;

		// Get the size (including the terminating null) of
		// the endpoint ID string of the waveOut device.
		mmr = waveOutMessage((HWAVEOUT)IntToPtr(waveOutId),
							 DRV_QUERYFUNCTIONINSTANCEIDSIZE,
							 (DWORD_PTR)&cbEndpointId, NULL);

		if (mmr != MMSYSERR_NOERROR ||
			cbEndpointIdKey != cbEndpointId)  // do sizes match?
		{
			continue;  // not a matching device
		}

		// Get the endpoint ID string for this waveOut device.
		mmr = waveOutMessage((HWAVEOUT)IntToPtr(waveOutId),
							 DRV_QUERYFUNCTIONINSTANCEID,
							 (DWORD_PTR)pstrEndpointId,
							 cbEndpointId);

		if (mmr != MMSYSERR_NOERROR)
		{
			continue;
		}

		// Check whether the endpoint ID string of this waveOut
		// device matches that of the audio endpoint device.
		if (lstrcmpiW(pstrEndpointId, pstrEndpointIdKey) == 0)
		{
			nDeviceID = waveOutId;  // found match
			break;
		}
	}

	if (waveOutId == cWaveOutDevices)
	{
		// We reached the end of the for-loop above without
		// finding a waveOut device with a matching endpoint
		// ID string. This behavior is quite unexpected.
		hr = E_UNEXPECTED;
	}

Exit:
    SAFE_RELEASE(pEnumerator);
    SAFE_RELEASE(pDevice);
    CoTaskMemFree(pstrEndpointIdKey);  // NULL pointer okay
    CoTaskMemFree(pstrEndpointId);
    return (unsigned int) nDeviceID;
}
}